{*******************************************************}
{                                                       }
{  idCGIRunner component for Internet Direct (Indy)     }
{                   HTTP Server                         }
{                                                       }
{ Author:                                               }
{ Serhiy Perevoznyk                                     }
{ serge_perevoznyk@hotmail.com                          }
{                                                       }
{     Use, modification and distribution is allowed     }
{without limitation, warranty, or liability of any kind.}
{                                                       }
{*******************************************************}

unit idCGIRunner;

interface

uses
  Windows,
  SysUtils,
  Classes,
  idTCPServer,
  idHTTPServer;

type
  TidCGIRunner = class(TComponent)
   private
     FServer : TidHTTPServer;
     FTimeOut : integer;
     FTimeOutMsg : string;
     FErrorMsg : string;
     FServerAdmin : string;
     FDocumentRoot : string;
     procedure SetServer(const AValue : TIdHTTPServer);
   protected
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
   public
     constructor Create(AOwner : TComponent); override;
     function Execute(LocalDoc : string;
                      AThread: TIdPeerThread;
                      RequestInfo: TIdHTTPRequestInfo;
                      ResponseInfo: TIdHTTPResponseInfo;
                      Action : string = '') : integer;
   published
     property Server : TidHTTPServer read FServer write SetServer;
     property TimeOut: integer read FTimeOut write FTimeOut default 5000;
     property TimeOutMsg : string read FTimeOutMsg write FTimeOutMsg;
     property ErrorMsg : string read FErrorMsg write FErrorMsg;
     property ServerAdmin : string read FServerAdmin write FServerAdmin;
     property DocumentRoot : string read FDocumentRoot write FDocumentRoot;
   end;

procedure Register;

implementation

type
  ThreadParams = record
    hReadPipe : THandle;
    s : String;
  end;
  PThreadParams = ^ThreadParams;

function ThreadRead(Params : Pointer):Dword; stdcall;
var
  Info : PThreadParams;
  Buffer :  array [0..4095] of Char;
  nb: DWord;
  i: Longint;
begin
  Result := 0;
  Info := PThreadParams(Params);
  while ReadFile( Info^.hReadPipe,
                  buffer,
                  SizeOf(buffer),
                  nb,
                  nil) do
     begin
       if nb = 0 then
         Break;
       for i:=0 to nb-1 do
         Info^.s := Info^.s + buffer[i];
      end;

end;


function GetEnv(const Name: string): string;
var
  Buffer: array[0..4095] of Char;
begin
  SetString(Result, Buffer, GetEnvironmentVariable(PChar(Name), Buffer, SizeOf(Buffer)));
end;



constructor TidCGIRunner.Create(AOwner: TComponent);
begin
  inherited;
  FErrorMsg := '<html><body><h1><center>Internal Server Error</body></html>'#13#10;
  FTimeOutMsg := '<html><body><h1><center>Process is terminated.</body></html>'#13#10;
  FServerAdmin := 'admin@server';
  FTimeOut := 5000;
end;

function TidCGIRunner.Execute(LocalDoc : string; AThread: TIdPeerThread;
                              RequestInfo: TIdHTTPRequestInfo;
                              ResponseInfo: TIdHTTPResponseInfo;
                              Action : string = '') : integer;
const
  FEnv = '%s=%s'#0;

var
  ParsLine : string;
  ParsList : TStringList;
  I        : integer;
  hReadPipe,
  hNIReadPipe,
  hWritePipe: THandle;
  saPipe: TSecurityAttributes;
  StartInfo: TStartupInfo;
  ProcInfo: TProcessInformation;
  Params : PThreadParams;
  ReaderID : Dword;
  ReaderHandle : THandle;
  cEnv : String;

begin
  cEnv := '';
  if Assigned(FServer) then
   cEnv :=        Format(FEnv,['SERVER_SOFTWARE', FServer.ServerSoftware ]);
  cEnv := cEnv + Format(FEnv,['SERVER_NAME',RequestInfo.Host]);;
  cEnv := cEnv + Format(FEnv,['SERVER_PROTOCOL',RequestInfo.Version]);
  cEnv := cEnv + Format(FEnv,['SERVER_PORT',IntToStr(AThread.Connection.Binding.Port)]);

  cEnv := cEnv + Format(FEnv,['GATEWAY_INTERFACE','CGI/1.1']);
  cEnv := cEnv + Format(FEnv,['REQUEST_METHOD',RequestInfo.Command]);
  cEnv := cEnv + Format(FEnv,['SCRIPT_NAME',RequestInfo.Document]);
  cEnv := cEnv + Format(FEnv,['QUERY_STRING',RequestInfo.UnparsedParams]);

  cEnv := cEnv + Format(FEnv,['REMOTE_ADDR',RequestInfo.RemoteIP]);

  //Win32 fields
  cEnv := cEnv + Format(FEnv,['SystemRoot',getenv('SystemRoot')]);
  cEnv := cEnv + Format(FEnv,['COMSPEC',getenv('COMSPEC')]);
  cEnv := cEnv + Format(FEnv,['WINDIR',getenv('WINDIR')]);
  cEnv := cEnv + Format(FEnv,['PATH',getenv('PATH')]);
  cEnv := cEnv + Format(FEnv,['PATH_INFO', Action]);
  cEnv := cEnv + Format(FEnv,['PATH_TRANSLATED',ExpandFileName(DocumentRoot + Action)]);

  //Add HTTP_ fields
  cEnv := cEnv + Format(FEnv,['HTTP_ACCEPT','*/*']);

  //Apache fields
  cEnv := cEnv + Format(FEnv,['SERVER_ADDR',AThread.Connection.Binding.IP]);
  cEnv := cEnv + Format(FEnv,['DOCUMENT_ROOT',DocumentRoot]);
  cEnv := cEnv + Format(FEnv,['SERVER_ADMIN',ServerAdmin]);
  cEnv := cEnv + Format(FEnv,['SCRIPT_FILENAME','']);

  cEnv := cEnv + #0;

  Params := nil;
  ReaderHandle := 0;
  saPipe.bInheritHandle := True;
  saPipe.lpSecurityDescriptor := nil;
  saPipe.nLength := SizeOf(saPipe);

  if not CreatePipe(hReadPipe,hWritePipe,@saPipe,0) then
    RaiseLastWin32Error;

  FillChar(StartInfo,SizeOf(StartInfo),0);

  StartInfo.dwFlags:= STARTF_USESTDHANDLES  or
                      STARTF_USESHOWWINDOW;
  StartInfo.wShowWindow:=SW_HIDE;
  StartInfo.hStdOutput := hWritePipe;
  StartInfo.hStdError  := hWritePipe;
  StartInfo.cb:=SizeOf(StartInfo);


try
  if not DuplicateHandle( GetCurrentProcess,
                          hReadPipe,
                          GetCurrentProcess,
                          @hNIReadPipe,
                          0,
                          False,
                          DUPLICATE_SAME_ACCESS) then
       begin
        ResponseInfo.ContentText := FErrorMsg;
        ResponseInfo.ContentLength := Length(ResponseInfo.ContentText);
        ResponseInfo.ResponseNo := 500;
        Result :=  ResponseInfo.ContentLength;
        Exit;
       end;

  CloseHandle(hReadPipe);

  New(Params);
  Params^.hReadPipe := hNIReadPipe;


  ReaderHandle := CreateThread( nil,
                                0,
                                @ThreadRead,
                                Params,
                                0,
                                ReaderId);
  if ReaderHandle = 0 then
   begin
     ResponseInfo.ContentText := FTimeOutMsg;
     ResponseInfo.ContentLength := Length(ResponseInfo.ContentText);
     ResponseInfo.ResponseNo := 500;
     Result :=  ResponseInfo.ContentLength;
     Exit;
   end;


  if CreateProcess(nil,
                   PChar(LocalDoc),
                   nil,
                   nil,
                   True,
                   0,
                   PChar(cEnv),
                   nil,
                   StartInfo,
                   ProcInfo)
  then
    begin
      CloseHandle(ProcInfo.hThread);
      CloseHandle(hWritePipe);
    end
  else
   begin
     ResponseInfo.ContentText := FTimeOutMsg;
     ResponseInfo.ContentLength := Length(ResponseInfo.ContentText);
     ResponseInfo.ResponseNo := 500;
     Result :=  ResponseInfo.ContentLength;
     Exit;
   end;


  if WaitForSingleObject(ReaderHandle, FTimeOUT) = WAIT_TIMEOUT then

     begin
        TerminateThread(ReaderHandle,0);
        TerminateProcess(ProcInfo.hProcess, 1);
        ResponseInfo.ContentText := FTimeOutMsg;
        ResponseInfo.ContentLength := Length(ResponseInfo.ContentText);
        ResponseInfo.ResponseNo := 500;
        Result :=  ResponseInfo.ContentLength;
        Exit;
     end;


   ParsLine := Params^.s;
   I := Pos(#13#10#13#10,Params^.s);
   if i > 0 then
    begin
      ParsLine := Copy(ParsLine,1,I+3);
      ParsLine := StringReplace(ParsLine,': ','=',[rfReplaceAll]);
      ParsList := TStringList.Create;
      ParsList.Text := ParsLine;
      ResponseInfo.ContentType   := ParsList.Values['content-type'];
      ResponseInfo.ResponseNo    := StrToIntDef(Copy(ParsList.Values['status'],1,3),200);
      ResponseInfo.WriteHeader;
      ResponseInfo.ContentText   := Copy(Params^.s,I+4, Length(Params^.s));
      ParsList.Free;
     end
       else
         begin
           ResponseInfo.ContentType := 'text/html';
         end;

   Result :=  ResponseInfo.ContentLength;

  finally
    if ReaderHandle > 0 then
       CloseHandle(ReaderHandle);
    if ProcInfo.hProcess > 0 then
       CloseHandle(ProcInfo.hProcess);
    if hReadPipe > 0 then
       CloseHandle(hNIReadPipe);
    if Assigned(Params) then
       Dispose(Params);
  end;

end;


procedure TidCGIRunner.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FServer) then
    FServer := nil;
end;

procedure TidCGIRunner.SetServer(const AValue: TIdHTTPServer);
begin
  if FServer <> AValue then
  begin
    if Avalue <> nil then AValue.FreeNotification(Self);
    FServer := AValue;
  end;
end;


procedure Register;
begin
  RegisterComponents('Indy Misc', [TidCGIRunner]);
end;

end.
